import requests
from requests_oauthlib import OAuth2Session
from oauthlib.oauth2 import BackendApplicationClient
from collections.abc import Callable
from typing import Optional, Type
from time import sleep
from urllib.parse import quote_plus
from json import loads
from secrets import token_urlsafe


from .cache.base import BaseCache
from .api import oauth
from .utils.utils import auth_host, api_host, slugify


class BattlenetClient:
    """ Creates a client to access the Battle.net REST API

    Attributes:
        oauth (:obj: `Oauth2Session`): the OAuth session to be used
        state (str, optional): saves the state for client verification
        region (str): the abbreviation of the region
        client_id (str): Client ID for use with the API
        client_secret (str): Client secret for use with the API
        redirect_uri (str, optional): Redirect URI for use with the API
        cache (:obj: `BaseCache`, optional): the cache to use
        config_endpoint (str, optional): endpoint to use for configuration when using OpenID Connection
        jwks (str, optional): endpoint to the JWKS keys
        token (dict): token used for authentication to API
    """
    oauth = None
    state = None
    region = None
    client_id = None
    client_secret = None
    redirect_uri = None
    cache = None
    config_endpoint = None
    jwks = None
    token = None

    def __init__(self, region: str, oauth_session: OAuth2Session, cache = Optional[Type[BaseCache]]):
        self.region = region
        self.oauth = oauth_session
        self.cache = cache
        self.auth_host = auth_host(self.region)
        self.api_host = api_host(self.region)

        if self.grant_type == 'client_credentials':
            self.fetch_token()

    @classmethod
    def authorization_code(cls, region, client_id: str, client_secret: str, redirect_uri: str, scope: list[str], *,
                           auto_refresh_url: Optional[str] = None, auto_refresh_kwargs: Optional[dict] = None,
                           updater: Callable[[dict], None] = None, cache: Optional[BaseCache] = None, ):
        """ Creates an instance of `Battlenet Client` using the Authorization Code Grant, or the Web Application Flow

        Args:
            region (str): The region where the client is connected
            client_id (str): Client ID issued by develop.battle.net
            client_secret (str): Client secret issued by develop.battle.net
            redirect_uri (str): URL to send to authorization service to return
            scope (list of str): list of scopes to grant
            auto_refresh_url (str, optional): Auto-refresh URL when token expires
            auto_refresh_kwargs (dict, optional): Extra arguments passed to OAuth2Session for automatic token refresh
            updater (function, optional): the function to perform the automatic update
            cache (:obj:`BaseCache`, optional): the cache to use

        Returns:
            :obj:`BattlenetClient` configured for Authorization Code Grant, or the Web Application Flow
        """
        cls.client_id = quote_plus(client_id)
        cls.client_secret = quote_plus(client_secret)
        cls.scope = ' '.join(scope)

        cls.grant_type = 'authorization_code'

        cls.redirect_uri = redirect_uri
        updater = updater if updater else cls.token_updater

        oauth_session = OAuth2Session(client_id, redirect_uri=redirect_uri, scope=scope, auto_refresh_url=auto_refresh_url,
                                      auto_refresh_kwargs=auto_refresh_kwargs, token_updater=updater)
        return cls(region, oauth_session, cache)

    @classmethod
    def client_credential(cls, region: str,  client_id: str, client_secret: str, *,
                          scope: Optional[list[str]] = None, auto_refresh_url: Optional[str] = None,
                          auto_refresh_kwargs: Optional[dict] = None, updater:Optional[Callable]=None,
                          cache = Optional[BaseCache]):
        """ Creates an instance of `Battlenet Client` using the Client Credentials Grant, or the Backend Application
        Flow

        Args:
            region (str): The region where the client is connected
            client_id (str): Client ID issued by develop.battle.net
            client_secret (str): Client secret issued by develop.battle.net
            scope (str): The scope of the request
            auto_refresh_url (str, optional): Auto-refresh URL when token expires
            auto_refresh_kwargs (dict, optional): Extra arguments passed to OAuth2Session for automatic token refresh
            updater (function, optional): specify the function to call when update to the token is required
            cache (:obj:`BaseCache`): cache system

        Returns:
            :obj: `BattlenetClient` configured for client credentials grant oauth2 session
        """
        cls.client_id = quote_plus(client_id)
        cls.client_secret = quote_plus(client_secret)
        cls.grant_type = 'client_credentials'

        cls.scope = " ".join(scope) if scope else None

        updater = cls.token_updater if updater is None else updater

        oauth_session = OAuth2Session(client=BackendApplicationClient(client_id=client_id),
                                      auto_refresh_url=auto_refresh_url,
                                      auto_refresh_kwargs=auto_refresh_kwargs,
                                      token_updater=updater)

        return cls(region, oauth_session, cache)

    @classmethod
    def open_id(cls, region, client_id: str, client_secret: str, redirect_uri: str, scope: list[str], *,
                           auto_refresh_url: Optional[str] = None, auto_refresh_kwargs: Optional[dict] = None,
                           updater: Callable[[dict], None] = None, cache: Optional[BaseCache] = None):
        """ Creates an instance of `Battlenet Client` using Open ID Connect

        Args:
            region (str): The region where the client is connected
            client_id (str): Client ID issued by develop.battle.net
            client_secret (str): Client secret issued by develop.battle.net
            redirect_uri (str): URL to send to authorization service to return
            scope (list of str): list of scopes to grant
            auto_refresh_url (str, optional): Auto-refresh URL when token expires
            auto_refresh_kwargs (dict, optional): Extra arguments passed to OAuth2Session for automatic token refresh
            updater (callable, optional): the function to perform the automatic update
            cache (:obj:`BaseCache`): cache system

        Returns:
            :obj:`BattlenetClient` configured for authorization code grant oauth2 session using Open ID Connect
        """
        cls.scope = scope.append("openid")
        cls.config_endpoint = f"{auth_host(region)}/.well-known/openid-configuration"
        cls.jwks = f"{auth_host(region)}/jwks/certs"

        return cls.authorization_code(region, client_id, client_secret, redirect_uri, scope,
                                      auto_refresh_url=auto_refresh_url, auto_refresh_kwargs=auto_refresh_kwargs,
                                      updater=updater, cache=cache)

    def get_authorization_url(self) -> str:
        """ Returns the authorization URL for authorization Authorization Code Grant, or the Web Application Flow

        Returns:
          str: authorization url
        """
        self.state = token_urlsafe(32)
        return oauth.authorization(self.region, self.client_id, self.redirect_uri, self.scope, self.state)

    def callback(self, code: str, state: str) -> bool:
        """ Processes the callback response from the authorization code grant.

        Args:
            code (str): code provided during the callback request
            state (str): state provided by the callback request
        """
        if state == self.state:
            self.fetch_token(code)
            return True

        return False

    def token_updater(self, token: dict):
        """ Updates the token data provided by `token`

        Args:
            token (dict): token returned from authorization server
        """
        if token:
            self.token = token


    def fetch_token(self, code: Optional[str]=None) -> None:
        """ Fetches the token from the authorization server

        Args
            code (str, optional): authorization code provided by authorization server
        """

        token = self.post(oauth.fetch_token,self.grant_type, redirect_uri=self.redirect_uri, code=code)

        if token:
            self.token_updater(loads(token))

    def is_token_valid(self) -> bool:
        """ Verifies that a given bearer token is valid and retrieves metadata about the token, including the client_id
        used to create the token, expiration timestamp, and scopes granted to the token.

        Returns:
            bool: True if the token is valid, False otherwise
        """
        return self.post(oauth.token_validation, self.oauth.access_token) is not None

    def get_user_info(self) -> Optional[dict]:
        """ Retrieves user info from the authorization server

        Returns:
            dict: user info
        """
        return self.get(oauth.user_info, "no-cache", self.oauth.access_token)

    def get(self, api_endpoint: Callable, category_name: str, *args, **kwargs) -> Optional[bytes]:
        """ Retrieves data from the API endpoint via HTTP GET or cache

        Args:
            api_endpoint (function): API endpoint function from the respective submodule
            category_name (str): category name from the respective submodule
            *args: list of positional arguments for the api endpoint function
            **kwargs: dict of keyword arguments for the api endpoint function

        Returns:
            bytes: data returned from the API endpoint if successful
            None: on failure
        """

        if self.cache:
            if self.cache.check(slugify(category_name), *args, **kwargs):
                return self.cache.select(slugify(category_name), *args, **kwargs)

        url, params, headers, data = api_endpoint(self.region, *args, **kwargs)

        if self.token:
            headers["Authorization"] = f"Bearer {self.token["access_token"]}"

        for _ in range(5):
            try:
                response = requests.get(url, auth=(self.client_id, self.client_secret), params=params, data=data,
                                        headers=headers)
                response.raise_for_status()
            except requests.HTTPError as error:
                if error.response.status_code == 429:
                    sleep(1)
                    continue
                if error.response.status_code in (400, 401):
                    return None
            else:
                if response.status_code == 200:
                    if self.cache:
                        self.cache.upsert(slugify(category_name), *args, data=response.content, **kwargs)

                    return response.content
        return None

    def post(self, api_endpoint,  *args, **kwargs) -> Optional[bytes]:
        """ Retrieves data from the API endpoint via HTTP GET or cache

        Args:
            api_endpoint (function): API endpoint function from the respective submodule
            *args: list of positional arguments for the api endpoint function
            **kwargs: dict of keyword arguments for the api endpoint function

        Returns:
            bytes: data returned from the API endpoint if successful
            None: on failure
        """

        url, params, headers, data = api_endpoint(self.region, *args, **kwargs)

        for _ in range(5):
            try:
                if self.token:
                    headers["Authorization"] = f"Bearer {self.token["access_token"]}"
                    response = requests.post(url, params=params, data=data, headers=headers)
                else:
                    response = requests.post(url, auth=(self.client_id, self.client_secret), params=params,
                                             headers=headers)

                response.raise_for_status()
            except requests.HTTPError as error:
                if error.response.status_code == 429:
                    sleep(1)
                    continue
                if error.response.status_code in (400, 401):
                    return None
            else:
                if response.status_code == 200:
                    return response.content
        else:
            return None

    def add_cache(self, cache):
        """ Associates the provided cache with the client

        Args:
            cache (:obj:`BaseCache`): cache system
        """
        self.cache = cache

    def close_cache(self):
        """ Closes the client assoicated cache """
        if self.cache:
            del self.cache


import pytest

from src.battlenet_client.utils.constants import VALID_REGIONS
from src.battlenet_client.api.oauth import user_info, token_validation
from src.battlenet_client.utils.exceptions import BNetRegionNotFoundError

from tests.constants import INVALID_REGIONS


@pytest.mark.parametrize('region_tag', VALID_REGIONS)
def test_user_info(region_tag):
    token = 'my_good_token_1234'
    data = user_info(region_tag, token)
    assert isinstance(data, tuple)
    assert isinstance(data[0], str)
    if region_tag == 'cn':
        assert data[0] == "https://oauth.battlenet.com.cn/userinfo"
    else:
        assert data[0] == f'https://oauth.battle.net/userinfo'



@pytest.mark.parametrize('region_tag', INVALID_REGIONS)
def test_user_info_invalid_region(region_tag):
    with pytest.raises(BNetRegionNotFoundError):
        user_info(region_tag)


@pytest.mark.parametrize('region_tag', VALID_REGIONS)
def test_token_validation(region_tag):
    token = 'my_good_token_1234'
    data = token_validation(region_tag, token)
    assert isinstance(data, tuple)
    assert isinstance(data[0], str)
    if region_tag == 'cn':
        assert data[0] == "https://oauth.battlenet.com.cn/check_token"
    else:
        assert data[0] == 'https://oauth.battle.net/check_token'
    assert data[1] is None
    assert isinstance(data[2], dict)
    assert 'token' in data[2].keys()
    assert token == data[2]["token"]
    assert data[3] is None


@pytest.mark.parametrize('region_tag', INVALID_REGIONS)
def test_token_validation_invalid_region(region_tag):
    with pytest.raises(BNetRegionNotFoundError):
        token = 'my_good_token_1234'
        token_validation(region_tag, token)



"""Defines the functions that generate the URLs, parameter lists, and headers
for the OAuth API Endpoints

Functions:
    user_info(region_tag, locale)
    validate_token(region_tag, token, locale)

Misc Variables:
    __version__
    __author__

Author: David "Gahd" Couples
License: GPL v3
Copyright: February 24, 2022
"""

from urllib.parse import urlencode
from typing import Tuple, Optional

from ..utils.decorators import verify_region
from ..utils.utils import auth_host

__version__ = "1.0.3"
__author__ = "David \"Gahd\" Couples"


@verify_region
def authorization(region_tag: str, client_id: str, redirect_uri: str, scope: str,
                  state: str, response_type: Optional[str]="code") -> str:
    """Required for authorization code flow

    Args:
        state (str): value suppplied to maintain state between request and callback
        scope (str): list of scopes to authorize
        redirect_uri (str): callback url
        client_id (str): client id provided by API access
        region_tag (str): region of the data request
        response_type (str, optional): response type (typically `code`)

    Returns:
        str: returns the URI with params encoded
    """
    params = {"response_type": response_type, "client_id": client_id, "redirect_uri": redirect_uri, "scope": scope,
              "state": state}

    return f"{auth_host(region_tag)}/authorize?{urlencode(params)}"


@verify_region
def fetch_token(region_tag: str, grant_type: str, *, code: Optional[str]=None,
          redirect_uri: Optional[str]=None) -> Tuple:
    """ Retrieves the token from a authorization code flow

    Args:
        region_tag (str): region of the data request
        grant_type (str): grant type , typically `authorization_code`
        code (str): authorization code received from the authorization server
        redirect_uri (str): return URL, needs to match the original request

    Returns:
        tuple(str, dict): the url, params, and data for retrieving the token
    """

    data = {"grant_type": grant_type}

    if grant_type != "client_credentials":
        data["code"] = code
        data["redirect_uri"] = redirect_uri

    return f"{auth_host(region_tag)}/token", None, None, data


@verify_region
def user_info(region_tag: str, token: str) -> Tuple:
    """Returns the user info

    Args:
        region_tag (str): region_tag abbreviation
        token (str): token provided by the authorization server

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    return  f"{auth_host(region_tag)}/userinfo", None, None, None


@verify_region
def token_validation(region_tag: str,  token: str) -> Tuple:
    """Returns if the token is still valid or not

    Args:
        region_tag (str): region_tag abbreviation
        token (str): token string to validate

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    return f"{auth_host(region_tag)}/check_token", None, None, {'token': token}



"""Defines exceptions related to the Battle.net OAuth API wrappers

Disclaimer:
    All rights reserved, Blizzard is the intellectual property owner of Battle.net and any data
    retrieved from this API.
"""


class BNetError(Exception):
    """Base Error class for BattleNet Client"""
    pass


class BNetRegionNotFoundError(BNetError):
    """Error raised when an invalid Battle.net region is detected"""
    pass


class BNetClientError(BNetError):
    """Error raised if there is a mismatch with the client and API endpoints"""
    pass


class BNetValueError(BNetError):
    """Exception for when bad data is present"""
    pass


class BNetReleaseError(BNetError):
    """Exception raised when there is a problem regarding the release"""
    pass


class BNetGrantError(BNetError):
    """Exception raised when there is a problem regarding the grant type"""
    pass




"""Defines the Battle.net Regions

Classes:
    Region: class to hold region identifiers

Misc Variables:
    __version__
    __author__

Author: David "Gahd" Couples
License: GPL v3
Copyright: 2025-10-16
"""

__version__ = '1.0.0'
__author__ = 'David \'Gahd\' Couples'


class Region:

    VALID_REGIONS = ('us', 'kr', 'eu', 'tw', 'cn')

    NORTH_AMERICA = "us"
    KOREA = "kr"
    EUROPE = "eu"
    TAIWAN = "tw"
    CHINA = "cn"
    APAC = ("kr", "tw")

    class Id:

        NORTH_AMERICA = 1
        KOREA = 2
        EUROPE = 3
        TAIWAN = 4
        CHINA = 5
        APAC = (2, 4)
